
import numpy as np
import prince_cr.config as config

class SemiLagrangianSolver(object):
    """Contains routines to project spectra from shifted grids back to old grid"""

    def __init__(self, cr_grid):
        # remember energy grid to project back to
        self.grid = cr_grid
        self.widths = cr_grid.widths
        self.grid_log = np.log(cr_grid.grid)
        self.bins_log = np.log(cr_grid.bins)

    def get_shifted_state(self, conloss, state):
        newbins = self.grid.bins - conloss

        # get shifted state from gradient
        newwidths  = (newbins[1:] - newbins[:-1])
        gradient = newwidths / self.widths
        newstate = state / gradient
        newstate = np.where(newstate > 1e-250, newstate, 1e-250)
        newstate_log = np.log(newstate)

        # get new centers
        newcenters = (newbins[1:] + newbins[:-1])/2
        newgrid_log = np.log(newcenters)

        return newgrid_log, newstate_log

    def interpolate(self, conloss, state):
        "Uses linear interpolation to find the new state old grid"
        # get shifted state
        newgrid_log, newstate_log = self.get_shifted_state(conloss, state)
        
        # return interpolated value as newstate
        return np.exp(np.interp(self.grid_log, newgrid_log, newstate_log))

    def interpolate_gradient(self, conloss, state):
        """Uses a linear approximation arround x_i to find the new state old grid"""
        # get shifted state
        newgrid_log, newstate_log = self.get_shifted_state(conloss, state)
        
        # find second order finite difference at x_i
        # and use for a linear approximation arround x_i
        deriv = np.gradient(newstate_log, newgrid_log, edge_order=2)
        return np.exp(newstate_log + deriv * (self.grid_log - newgrid_log))

    def interpolate_linear_weights(self, conloss, state):
        """Uses linear interpolation with lagrange polynomials
        (same as interpolate, but directly implemented for testing)"""
        # get shifted state
        newgrid_log, newstate_log = self.get_shifted_state(conloss, state)
        
        # slices for indexing
        i0 = slice(None,-1)
        i1 = slice(1,None)
        # # calculate weights from old and new grid
        weight0 = (self.grid_log[i0] - newgrid_log[i1]) / (newgrid_log[i0] - newgrid_log[i1])
        weight1 = (self.grid_log[i0] - newgrid_log[i0]) / (newgrid_log[i1] - newgrid_log[i0])

        # calculate state by interpolating with weights
        state = np.zeros_like(self.grid_log)
        state[i0] = np.exp(weight0 * newstate_log[i0] + weight1 * newstate_log[i1])
        return state

    def interpolate_quadratic_weights(self, conloss, state):
        """Uses quadratic interpolation with lagrange polynomials"""
        # get shifted state
        newgrid_log, newstate_log = self.get_shifted_state(conloss, state)
        
        i0 = slice(None,-3)
        i1 = slice(1,-2)
        i2 = slice(2,-1)
        i3 = slice(3,None)

        # calculate backward weights from old and new grid
        weight_b0 = ((self.grid_log[i1] - newgrid_log[i1]) / (newgrid_log[i0] - newgrid_log[i1])
                    *(self.grid_log[i1] - newgrid_log[i2]) / (newgrid_log[i0] - newgrid_log[i2])
                    )
        weight_b1 = ((self.grid_log[i1] - newgrid_log[i0]) / (newgrid_log[i1] - newgrid_log[i0])
                    *(self.grid_log[i1] - newgrid_log[i2]) / (newgrid_log[i1] - newgrid_log[i2])
                    )
        weight_b2 = ((self.grid_log[i1] - newgrid_log[i0]) / (newgrid_log[i2] - newgrid_log[i0])
                    *(self.grid_log[i1] - newgrid_log[i1]) / (newgrid_log[i2] - newgrid_log[i1])
                    )

        weight_f1 = ((self.grid_log[i1] - newgrid_log[i2]) / (newgrid_log[i1] - newgrid_log[i2])
                    *(self.grid_log[i1] - newgrid_log[i3]) / (newgrid_log[i1] - newgrid_log[i3])
                    )
        weight_f2 = ((self.grid_log[i1] - newgrid_log[i1]) / (newgrid_log[i2] - newgrid_log[i1])
                    *(self.grid_log[i1] - newgrid_log[i3]) / (newgrid_log[i2] - newgrid_log[i3])
                    )
        weight_f3 = ((self.grid_log[i1] - newgrid_log[i1]) / (newgrid_log[i3] - newgrid_log[i1])
                    *(self.grid_log[i1] - newgrid_log[i2]) / (newgrid_log[i3] - newgrid_log[i2])
                    )

        # calculate state by interpolating with weights
        state = np.zeros_like(self.grid_log)
        # backward only
        # state[i1] = np.exp(weight_b0 * newstate_log[i0] + weight_b1 * newstate_log[i1] + weight_b2 * newstate_log[i2])
        # forward only
        # state[i1] = np.exp(weight_f1 * newstate_log[i1] + weight_f2 * newstate_log[i2] + weight_f3 * newstate_log[i3])

        # forward and backward average
        state[i1] = np.exp(weight_b0 / 2 * newstate_log[i0] + (weight_f1 + weight_b1) / 2 * newstate_log[i1]
                          + (weight_f2 + weight_b2) / 2 * newstate_log[i2] + weight_f3 / 2 * newstate_log[i3]
                          )
        return state

    def interpolate_cubic_weights(self, conloss, state):
        """Uses cubic interpolation with lagrange polynomials"""
        # get shifted state
        newgrid_log, newstate_log = self.get_shifted_state(conloss, state)
        
        i0 = slice(None,-3)
        i1 = slice(1,-2)
        i2 = slice(2,-1)
        i3 = slice(3,None)

        # calculate weights from old and new grid
        weight0 = ((self.grid_log[i1] - newgrid_log[i1]) / (newgrid_log[i0] - newgrid_log[i1])
                    *(self.grid_log[i1] - newgrid_log[i2]) / (newgrid_log[i0] - newgrid_log[i2])
                    *(self.grid_log[i1] - newgrid_log[i3]) / (newgrid_log[i0] - newgrid_log[i3])
                    )
        weight1 = ((self.grid_log[i1] - newgrid_log[i0]) / (newgrid_log[i1] - newgrid_log[i0])
                    *(self.grid_log[i1] - newgrid_log[i2]) / (newgrid_log[i1] - newgrid_log[i2])
                    *(self.grid_log[i1] - newgrid_log[i3]) / (newgrid_log[i1] - newgrid_log[i3])
                    )
        weight2 = ((self.grid_log[i1] - newgrid_log[i0]) / (newgrid_log[i2] - newgrid_log[i0])
                    *(self.grid_log[i1] - newgrid_log[i1]) / (newgrid_log[i2] - newgrid_log[i1])
                    *(self.grid_log[i1] - newgrid_log[i3]) / (newgrid_log[i2] - newgrid_log[i3])
                    )
        weight3 = ((self.grid_log[i1] - newgrid_log[i0]) / (newgrid_log[i3] - newgrid_log[i0])
                    *(self.grid_log[i1] - newgrid_log[i1]) / (newgrid_log[i3] - newgrid_log[i1])
                    *(self.grid_log[i1] - newgrid_log[i2]) / (newgrid_log[i3] - newgrid_log[i2])
                    )

        # reverse0 = ((newgrid_log[i2] - self.grid_log[i1]) / (self.grid_log[i0] - self.grid_log[i1])
        #             *(newgrid_log[i2] - self.grid_log[i2]) / (self.grid_log[i0] - self.grid_log[i2])
        #             *(newgrid_log[i2] - self.grid_log[i3]) / (self.grid_log[i0] - self.grid_log[i3])
        #             )
        # reverse1 = ((newgrid_log[i2] - self.grid_log[i0]) / (self.grid_log[i1] - self.grid_log[i0])
        #             *(newgrid_log[i2] - self.grid_log[i2]) / (self.grid_log[i1] - self.grid_log[i2])
        #             *(newgrid_log[i2] - self.grid_log[i3]) / (self.grid_log[i1] - self.grid_log[i3])
        #             )
        # reverse2 = ((newgrid_log[i2] - self.grid_log[i0]) / (self.grid_log[i2] - self.grid_log[i0])
        #             *(newgrid_log[i2] - self.grid_log[i1]) / (self.grid_log[i2] - self.grid_log[i1])
        #             *(newgrid_log[i2] - self.grid_log[i3]) / (self.grid_log[i2] - self.grid_log[i3])
        #             )
        # reverse3 = ((newgrid_log[i2] - self.grid_log[i0]) / (self.grid_log[i3] - self.grid_log[i0])
        #             *(newgrid_log[i2] - self.grid_log[i1]) / (self.grid_log[i3] - self.grid_log[i1])
        #             *(newgrid_log[i2] - self.grid_log[i2]) / (self.grid_log[i3] - self.grid_log[i2])
        #             )


        # weight_sum = weight0[i3] + weight1[i2] + weight2[i1] + weight3[i0]
        # weight_sum = weight0[i0] + weight1[i1] + weight2[i2] + weight3[i3]

        # if spid == 101:
        #     print (1 - weight_sum) 
        # weight0[i3] += (1 - weight_sum) * reverse0[i3]
        # weight1[i2] += (1 - weight_sum) * reverse1[i2]
        # weight2[i1] += (1 - weight_sum) * reverse2[i1]
        # weight3[i0] += (1 - weight_sum) * reverse3[i0]

        # weight0[i0] += (1 - weight_sum) * reverse0[i0]
        # weight1[i1] += (1 - weight_sum) * reverse1[i1]
        # weight2[i2] += (1 - weight_sum) * reverse2[i2]
        # weight3[i3] += (1 - weight_sum) * reverse3[i3]
        state = np.zeros_like(self.grid_log)
        state[i1] = np.exp( weight0 * newstate_log[i0] + weight1 * newstate_log[i1]
                          + weight2 * newstate_log[i2] + weight3 * newstate_log[i3]
                          )
        return state

    def interpolate_4thorder_weights(self, conloss, state):
        """Uses quadratic interpolation with lagrange polynomials"""
        # get shifted state
        newgrid_log, newstate_log = self.get_shifted_state(conloss, state)
        
        i0 = slice(None,-5)
        i1 = slice(1,-4)
        i2 = slice(2,-3)
        i3 = slice(3,-2)
        i4 = slice(4,-1)
        # i5 = slice(5,None)

        # calculate backward weights from old and new grid
        weight_b0 = ((self.grid_log[i2] - newgrid_log[i1]) / (newgrid_log[i0] - newgrid_log[i1])
                    *(self.grid_log[i2] - newgrid_log[i2]) / (newgrid_log[i0] - newgrid_log[i2])
                    *(self.grid_log[i2] - newgrid_log[i3]) / (newgrid_log[i0] - newgrid_log[i3])
                    *(self.grid_log[i2] - newgrid_log[i4]) / (newgrid_log[i0] - newgrid_log[i4])
                    )
        weight_b1 = ((self.grid_log[i2] - newgrid_log[i0]) / (newgrid_log[i1] - newgrid_log[i0])
                    *(self.grid_log[i2] - newgrid_log[i2]) / (newgrid_log[i1] - newgrid_log[i2])
                    *(self.grid_log[i2] - newgrid_log[i3]) / (newgrid_log[i1] - newgrid_log[i3])
                    *(self.grid_log[i2] - newgrid_log[i4]) / (newgrid_log[i1] - newgrid_log[i4])
                    )
        weight_b2 = ((self.grid_log[i2] - newgrid_log[i0]) / (newgrid_log[i2] - newgrid_log[i0])
                    *(self.grid_log[i2] - newgrid_log[i1]) / (newgrid_log[i2] - newgrid_log[i1])
                    *(self.grid_log[i2] - newgrid_log[i3]) / (newgrid_log[i2] - newgrid_log[i3])
                    *(self.grid_log[i2] - newgrid_log[i4]) / (newgrid_log[i2] - newgrid_log[i4])
                    )
        weight_b3 = ((self.grid_log[i2] - newgrid_log[i0]) / (newgrid_log[i3] - newgrid_log[i0])
                    *(self.grid_log[i2] - newgrid_log[i1]) / (newgrid_log[i3] - newgrid_log[i1])
                    *(self.grid_log[i2] - newgrid_log[i2]) / (newgrid_log[i3] - newgrid_log[i2])
                    *(self.grid_log[i2] - newgrid_log[i4]) / (newgrid_log[i3] - newgrid_log[i4])
                    )
        weight_b4 = ((self.grid_log[i2] - newgrid_log[i0]) / (newgrid_log[i4] - newgrid_log[i0])
                    *(self.grid_log[i2] - newgrid_log[i1]) / (newgrid_log[i4] - newgrid_log[i1])
                    *(self.grid_log[i2] - newgrid_log[i2]) / (newgrid_log[i4] - newgrid_log[i2])
                    *(self.grid_log[i2] - newgrid_log[i3]) / (newgrid_log[i4] - newgrid_log[i3])
                    )

        # weight_f1 = ((self.grid_log[i2] - newgrid_log[i2]) / (newgrid_log[i1] - newgrid_log[i2])
        #             *(self.grid_log[i2] - newgrid_log[i3]) / (newgrid_log[i1] - newgrid_log[i3])
        #             *(self.grid_log[i2] - newgrid_log[i4]) / (newgrid_log[i1] - newgrid_log[i4])
        #             *(self.grid_log[i2] - newgrid_log[i5]) / (newgrid_log[i1] - newgrid_log[i5])
        #             )
        # weight_f2 = ((self.grid_log[i2] - newgrid_log[i1]) / (newgrid_log[i2] - newgrid_log[i1])
        #             *(self.grid_log[i2] - newgrid_log[i3]) / (newgrid_log[i2] - newgrid_log[i3])
        #             *(self.grid_log[i2] - newgrid_log[i4]) / (newgrid_log[i2] - newgrid_log[i4])
        #             *(self.grid_log[i2] - newgrid_log[i5]) / (newgrid_log[i2] - newgrid_log[i5])
        #             )
        # weight_f3 = ((self.grid_log[i2] - newgrid_log[i1]) / (newgrid_log[i3] - newgrid_log[i1])
        #             *(self.grid_log[i2] - newgrid_log[i2]) / (newgrid_log[i3] - newgrid_log[i2])
        #             *(self.grid_log[i2] - newgrid_log[i4]) / (newgrid_log[i3] - newgrid_log[i4])
        #             *(self.grid_log[i2] - newgrid_log[i5]) / (newgrid_log[i3] - newgrid_log[i5])
        #             )
        # weight_f4 = ((self.grid_log[i2] - newgrid_log[i1]) / (newgrid_log[i4] - newgrid_log[i1])
        #             *(self.grid_log[i2] - newgrid_log[i2]) / (newgrid_log[i4] - newgrid_log[i2])
        #             *(self.grid_log[i2] - newgrid_log[i3]) / (newgrid_log[i4] - newgrid_log[i3])
        #             *(self.grid_log[i2] - newgrid_log[i5]) / (newgrid_log[i4] - newgrid_log[i5])
        #             )
        # weight_f5 = ((self.grid_log[i2] - newgrid_log[i1]) / (newgrid_log[i5] - newgrid_log[i1])
        #             *(self.grid_log[i2] - newgrid_log[i2]) / (newgrid_log[i5] - newgrid_log[i2])
        #             *(self.grid_log[i2] - newgrid_log[i3]) / (newgrid_log[i5] - newgrid_log[i3])
        #             *(self.grid_log[i2] - newgrid_log[i4]) / (newgrid_log[i5] - newgrid_log[i4])
        #             )

        # calculate state by interpolating with weights
        state = np.zeros_like(self.grid_log)
        # backward only
        state[i2] = np.exp(weight_b0 * newstate_log[i0] + weight_b1 * newstate_log[i1]
                   + weight_b2 * newstate_log[i2] + weight_b3 * newstate_log[i3] + weight_b4 * newstate_log[i4])
        # forward only
        # state[i2] = np.exp(weight_f1 * newstate_log[i1] + weight_f2 * newstate_log[i2]
        #            + weight_f3 * newstate_log[i3] + weight_f4 * newstate_log[i4] + weight_f5 * newstate_log[i5])
        # forward and backward average
        # state[i2] = np.exp(weight_b0 / 2 * newstate_log[i0] + (weight_f1 + weight_b1) / 2 * newstate_log[i1]
        #                   + (weight_f2 + weight_b2) / 2 * newstate_log[i2] + (weight_f3 + weight_b3) / 2 * newstate_log[i3] 
        #                   + (weight_f4 + weight_b4) / 2 * newstate_log[i4] + weight_f5 / 2 * newstate_log[i5]
        #                   )
        return state


    def interpolate_5thorder_weights(self, conloss, state):
        """Uses cubic interpolation with lagrange polynomials"""
        # get shifted state
        newgrid_log, newstate_log = self.get_shifted_state(conloss, state)
        
        i0 = slice(None,-5)
        i1 = slice(1,-4)
        i2 = slice(2,-3)
        i3 = slice(3,-2)
        i4 = slice(4,-1)
        i5 = slice(5,None)
        # calculate weights from old and new grid
        weight0 = ((self.grid_log[i2] - newgrid_log[i1]) / (newgrid_log[i0] - newgrid_log[i1])
                    *(self.grid_log[i2] - newgrid_log[i2]) / (newgrid_log[i0] - newgrid_log[i2])
                    *(self.grid_log[i2] - newgrid_log[i3]) / (newgrid_log[i0] - newgrid_log[i3])
                    *(self.grid_log[i2] - newgrid_log[i4]) / (newgrid_log[i0] - newgrid_log[i4])
                    *(self.grid_log[i2] - newgrid_log[i5]) / (newgrid_log[i0] - newgrid_log[i5])
                    )
        weight1 = ((self.grid_log[i2] - newgrid_log[i0]) / (newgrid_log[i1] - newgrid_log[i0])
                    *(self.grid_log[i2] - newgrid_log[i2]) / (newgrid_log[i1] - newgrid_log[i2])
                    *(self.grid_log[i2] - newgrid_log[i3]) / (newgrid_log[i1] - newgrid_log[i3])
                    *(self.grid_log[i2] - newgrid_log[i4]) / (newgrid_log[i1] - newgrid_log[i4])
                    *(self.grid_log[i2] - newgrid_log[i5]) / (newgrid_log[i1] - newgrid_log[i5])
                    )
        weight2 = ((self.grid_log[i2] - newgrid_log[i0]) / (newgrid_log[i2] - newgrid_log[i0])
                    *(self.grid_log[i2] - newgrid_log[i1]) / (newgrid_log[i2] - newgrid_log[i1])
                    *(self.grid_log[i2] - newgrid_log[i3]) / (newgrid_log[i2] - newgrid_log[i3])
                    *(self.grid_log[i2] - newgrid_log[i4]) / (newgrid_log[i2] - newgrid_log[i4])
                    *(self.grid_log[i2] - newgrid_log[i5]) / (newgrid_log[i2] - newgrid_log[i5])
                    )
        weight3 = ((self.grid_log[i2] - newgrid_log[i0]) / (newgrid_log[i3] - newgrid_log[i0])
                    *(self.grid_log[i2] - newgrid_log[i1]) / (newgrid_log[i3] - newgrid_log[i1])
                    *(self.grid_log[i2] - newgrid_log[i2]) / (newgrid_log[i3] - newgrid_log[i2])
                    *(self.grid_log[i2] - newgrid_log[i4]) / (newgrid_log[i3] - newgrid_log[i4])
                    *(self.grid_log[i2] - newgrid_log[i5]) / (newgrid_log[i3] - newgrid_log[i5])
                    )
        weight4 = ((self.grid_log[i2] - newgrid_log[i0]) / (newgrid_log[i4] - newgrid_log[i0])
                    *(self.grid_log[i2] - newgrid_log[i1]) / (newgrid_log[i4] - newgrid_log[i1])
                    *(self.grid_log[i2] - newgrid_log[i2]) / (newgrid_log[i4] - newgrid_log[i2])
                    *(self.grid_log[i2] - newgrid_log[i3]) / (newgrid_log[i4] - newgrid_log[i3])
                    *(self.grid_log[i2] - newgrid_log[i5]) / (newgrid_log[i4] - newgrid_log[i5])
                    )
        weight5 = ((self.grid_log[i2] - newgrid_log[i0]) / (newgrid_log[i5] - newgrid_log[i0])
                    *(self.grid_log[i2] - newgrid_log[i1]) / (newgrid_log[i5] - newgrid_log[i1])
                    *(self.grid_log[i2] - newgrid_log[i2]) / (newgrid_log[i5] - newgrid_log[i2])
                    *(self.grid_log[i2] - newgrid_log[i3]) / (newgrid_log[i5] - newgrid_log[i3])
                    *(self.grid_log[i2] - newgrid_log[i4]) / (newgrid_log[i5] - newgrid_log[i4])
                    )

        # calculate state by interpolating with weights
        state = np.zeros_like(self.grid_log)
        state[i2] = np.exp( weight0 * newstate_log[i0] + weight1 * newstate_log[i1]
                          + weight2 * newstate_log[i2] + weight3 * newstate_log[i3]
                          + weight4 * newstate_log[i4] + weight5 * newstate_log[i5]
                          )
        return state

class DifferentialOperator(object):
    def __init__(self, cr_grid, nspec):
        self.ebins = cr_grid.bins
        self.egrid = cr_grid.grid
        self.ewidths = cr_grid.widths
        self.dim_e = cr_grid.d
        # binsize in log(e)
        self.log_width = np.log(self.ebins[1] / self.ebins[0])

        self.nspec = nspec
        self.operator = self.construct_differential_operator()
        if config.linear_algebra_backend.lower() == 'cupy':
            import cupyx
            self.operator = cupyx.scipy.sparse.csr_matrix(self.operator)

    def construct_differential_operator(self):
        from scipy.sparse import coo_matrix, block_diag
        # # Construct a 
        # # First rows of operator matrix
        # diags_leftmost = [0, 1, 2, 3]
        # coeffs_leftmost = [-11, 18, -9, 2]
        # denom_leftmost = 6.
        # diags_left_1 = [-1, 0, 1, 2, 3]
        # coeffs_left_1 = [-3, -10, 18, -6, 1]
        # denom_left_1 = 12.
        # diags_left_2 = [-2, -1, 0, 1, 2, 3]
        # coeffs_left_2 = [3, -30, -20, 60, -15, 2]
        # denom_left_2 = 60.

        # # Centered diagonals
        # diags = [-3, -2, -1, 1, 2, 3]
        # coeffs = [-1, 9, -45, 45, -9, 1]
        # denom = 60.

        # # First rows of operator matrix
        # diags_leftmost = [0, 1]
        # coeffs_leftmost = [-1, 1]
        # denom_leftmost = 1.
        # diags_left_1 = [0, 1]
        # coeffs_left_1 = [-1, 1]
        # denom_left_1 = 1.
        # diags_left_2 = [0, 1]
        # coeffs_left_2 = [-1, 1]
        # denom_left_2 = 1.

        # # Centered diagonals
        # diags = [0, 1]
        # coeffs = [-1, 1]
        # denom = 1.


        # First rows of operator matrix
        diags_leftmost = [1, 2, 3]
        coeffs_leftmost = [-3, 4, -1]
        denom_leftmost = 2.
        diags_left_1 = [-1, 0, 1, 2]
        coeffs_left_1 = [-2, -3, 6, -1]
        denom_left_1 = 6.
        diags_left_2 = [-1, 0, 1, 2]
        coeffs_left_2 = [-2, -3, 6, -1]
        denom_left_2 = 6.

        # # Centered diagonals
        # diags = [-1, 0, 1, 2]
        # coeffs = [-2, -3, 6, -1]
        # denom = 6.

        


        # # First rows of operator matrix
        # diags_leftmost = [0, 1, 2, 3]
        # coeffs_leftmost = [-11, 18, -9, 2]
        # denom_leftmost = 6.
        # diags_left_1 = [-1, 0, 1, 2, 3]
        # coeffs_left_1 = [-3, -10, 18, -6, 1]
        # denom_left_1 = 12.
        # diags_left_2 = [-2, -1, 0, 1, 2, 3]
        # coeffs_left_2 = [3, -30, -20, 60, -15, 2]
        # denom_left_2 = 60.

        # # Centered diagonals
        # diags = [-2, -1, 0, 1, 2, 3]
        # coeffs = [3, -30, -20, 60, -15, 2]
        # denom = 60.


        # print diags
        # print coeffs
        # diags = [-d for d in diags[::-1]]
        # coeffs = [-d for d in coeffs[::-1]]
        # denom = denom
        # print diags
        # print coeffs
        # diags = [-3, -2, -1, 1, 2, 3]
        # coeffs = [-1, 9, -45, 45, -9, 1]
        # denom = 60.

        diags = [-1, 0, 1, 2, 3]
        coeffs = [-3, -10, 18, -6, 1]
        denom = 12.

        # diags = [0, 1, 2, 3]
        # coeffs = [-11, 18, -9, 2]
        # denom = 6.



        # diags_leftmost = [0, 1]
        # coeffs_leftmost = [-1, 1]
        # denom_leftmost = 1.
        # diags_left_1 = [0, 1]
        # coeffs_left_1 = [-1, 1]
        # denom_left_1 = 1.
        # diags_left_2 = [0, 1]
        # coeffs_left_2 = [-1, 1]
        # denom_left_2 = 1.
        # diags = [0, 1]
        # coeffs = [-1, 1]
        # denom = 1.

        # Last rows at the right of operator matrix
        diags_right_2 = [-d for d in diags_left_2[::-1]]
        coeffs_right_2 = [-d for d in coeffs_left_2[::-1]]
        denom_right_2 = denom_left_2
        diags_right_1 = [-d for d in diags_left_1[::-1]]
        coeffs_right_1 = [-d for d in coeffs_left_1[::-1]]
        denom_right_1 = denom_left_1
        diags_rightmost = [-d for d in diags_leftmost[::-1]]
        coeffs_rightmost = [-d for d in coeffs_leftmost[::-1]]
        denom_rightmost = denom_leftmost

        h = self.log_width
        dim_e = self.dim_e
        last = dim_e - 1

        op_matrix = np.zeros((dim_e, dim_e))
        op_matrix[0, np.asarray(diags_leftmost)] = np.asarray(
            coeffs_leftmost) / (denom_leftmost * h)
        op_matrix[1, 1 + np.asarray(diags_left_1)] = np.asarray(
            coeffs_left_1) / (denom_left_1 * h)
        op_matrix[2, 2 + np.asarray(diags_left_2)] = np.asarray(
            coeffs_left_2) / (denom_left_2 * h)
        op_matrix[last,
                  last + np.asarray(diags_rightmost)] = np.asarray(
                      coeffs_rightmost) / (denom_rightmost * h)
        op_matrix[last - 1, last - 1 + np.asarray(
            diags_right_1)] = np.asarray(coeffs_right_1) / (denom_right_1 * h)
        op_matrix[last - 2, last - 2 + np.asarray(
            diags_right_2)] = np.asarray(coeffs_right_2) / (denom_right_2 * h)
        for row in range(3, dim_e - 3):
            op_matrix[row, row + np.asarray(diags)] = np.asarray(coeffs) / (
                denom * h)
        # Construct an operator by left multiplication of the back-substitution
        # dlnE to dE. The right energy loss has to be later multiplied in every step
        single_op = coo_matrix(
            np.diag(1 / self.egrid).dot(op_matrix)
        )

        # construct the operator for the whole matrix, by repeating
        return block_diag(self.nspec*[single_op]).tocsr()


    def solve_coefficients(self,stencils,degree = 1):
        """Calculates the finite difference coefficients for given stencils.

        Note: The function sets up a linear equation system and solves it numerically
              Do not expect the result to be 100% accurate, as the coefficients are usually fractions

        Args:
            stencils (list of integers): position of stencils on regular grid
            degree (integer): degree of derviative, default: 1
        """
        if len(stencils) < degree + 1:
            raise Exception('Not enough stencils to solve for dervative of degree {:}, stencils given: {}'.format(degree,stencils))

        # setup of equation system
        exponents = np.arange(len(stencils))
        matrix = np.power.outer(stencils,exponents).T # pylint: disable=no-member
        right = np.zeros_like(stencils)
        right[degree] = np.math.factorial(degree)

        # solution
        return  np.linalg.solve(matrix,right)



# Test whether cross section are correctly created.

import numpy as np

import prince_cr.config as config
from prince_cr import cross_sections, photonfields, core

config.x_cut = 1e-4
config.x_cut_proton = 1e-2
config.tau_dec_threshold = np.inf
config.max_mass = 14
config.debug_level = 0 # suppress info statements

pf = photonfields.CombinedPhotonField(
                [photonfields.CMBPhotonSpectrum, 
                photonfields.CIBGilmore2D])
cs = cross_sections.CompositeCrossSection([(0., cross_sections.TabulatedCrossSection, ('CRP2_TALYS',)),
                                        (0.14, cross_sections.SophiaSuperposition, ())])

import unittest
class TestCsec(unittest.TestCase):

    def test_kernel_1(self):
        prince_run_talys = core.PriNCeRun(max_mass = 4, photon_field=pf, cross_sections=cs)
        self.assertEqual(prince_run_talys.int_rates._batch_matrix.shape, (88344, 72))
        self.assertEqual(prince_run_talys.int_rates._batch_rows.shape, (88344,))
        self.assertEqual(prince_run_talys.int_rates._batch_cols.shape, (88344,))
        self.assertEqual(prince_run_talys.int_rates._batch_vec.shape, (88344,))

    def test_kernel_4(self):
        prince_run_talys = core.PriNCeRun(max_mass = 1, photon_field=pf, cross_sections=cs)
        self.assertEqual(prince_run_talys.int_rates._batch_matrix.shape, (17528, 72))
        self.assertEqual(prince_run_talys.int_rates._batch_rows.shape, (17528,))
        self.assertEqual(prince_run_talys.int_rates._batch_cols.shape, (17528,))
        self.assertEqual(prince_run_talys.int_rates._batch_vec.shape, (17528,))

    def test_kernel_14(self):
        prince_run_talys = core.PriNCeRun(max_mass = 14, photon_field=pf, cross_sections=cs)
        self.assertEqual(prince_run_talys.int_rates._batch_matrix.shape, (287664, 72))
        self.assertEqual(prince_run_talys.int_rates._batch_rows.shape, (287664,))
        self.assertEqual(prince_run_talys.int_rates._batch_cols.shape, (287664,))
        self.assertEqual(prince_run_talys.int_rates._batch_vec.shape, (287664,))

if __name__ == '__main__':
    unittest.main()


'''Provides user interface and runtime management.'''

import pickle as pickle
from os import path

import numpy as np

from prince_cr import cross_sections, data, interaction_rates
from prince_cr.data import EnergyGrid
from prince_cr.util import info, get_AZN
import prince_cr.config as config


class PriNCeRun(object):
    """This is a draft of the main class.

    This class is supposed to interprete the config options
    and initialize all necessary stuff in order. This class is
    meant to keep all separate ingredients together in one place,
    and it is inteded to be passed to further classes via `self`.
    """

    def __init__(self, *args, **kwargs):

        if "max_mass" in kwargs:
            max_mass = kwargs.pop("max_mass", config.max_mass)

        # Initialize energy grid
        if config.grid_scale == 'E':
            info(1, 'initialising Energy grid')
            self.cr_grid = EnergyGrid(*config.cosmic_ray_grid)
            self.ph_grid = EnergyGrid(*config.photon_grid)
        else:
            raise Exception(
                "Unknown energy grid scale {:}, adjust config.grid_scale".
                format(config.grid_scale))

        # Cross section handler
        if 'cross_sections' in kwargs:
            self.cross_sections = kwargs['cross_sections']
        else:
            self.cross_sections = cross_sections.CompositeCrossSection(
                [(0., cross_sections.TabulatedCrossSection, ('CRP2_TALYS', )),
                 (0.14, cross_sections.SophiaSuperposition, ())])

        # Photon field handler
        if 'photon_field' in kwargs:
            self.photon_field = kwargs['photon_field']
        else:
            import prince_cr.photonfields as pf
            self.photon_field = pf.CombinedPhotonField(
                [pf.CMBPhotonSpectrum, pf.CIBGilmore2D])

        # Limit max nuclear mass of eqn system
        if "species_list" in kwargs:
            system_species = list(
                set(kwargs["species_list"]) & set(
                    self.cross_sections.known_species))
        else:
            system_species = [
                s for s in self.cross_sections.known_species
                if get_AZN(s)[0] <= max_mass
            ]
        # Disable photo-meson production
        if not config.secondaries:
            system_species = [s for s in system_species if s >= 100]
        # Remove particles that are explicitly excluded
        for pid in config.ignore_particles:
            if pid in system_species:
                system_species.remove(pid)

        # Initialize species manager for all species for which cross sections are known
        self.spec_man = data.SpeciesManager(system_species, self.cr_grid.d)

        # Total dimension of system
        self.dim_states = self.cr_grid.d * self.spec_man.nspec
        self.dim_bins = (self.cr_grid.d + 1) * self.spec_man.nspec

        # Initialize continuous energy losses
        self.adia_loss_rates_grid = interaction_rates.ContinuousAdiabaticLossRate(
            prince_run=self, energy='grid')
        self.pair_loss_rates_grid = interaction_rates.ContinuousPairProductionLossRate(
            prince_run=self, energy='grid')
        self.adia_loss_rates_bins = interaction_rates.ContinuousAdiabaticLossRate(
            prince_run=self, energy='bins')
        self.pair_loss_rates_bins = interaction_rates.ContinuousPairProductionLossRate(
            prince_run=self, energy='bins')

        # Initialize the interaction rates
        self.int_rates = interaction_rates.PhotoNuclearInteractionRate(
            prince_run=self)

        # Let species manager know about the photon grid dimensions (for idx calculations)
        # it is accesible under index "ph" for lidx(), uidx() calls
        self.spec_man.add_grid('ph', self.ph_grid.d)


    def set_photon_field(self, pfield):
        self.photon_field = pfield
        self.adia_loss_rates_grid.photon_field = pfield
        self.pair_loss_rates_grid.photon_field = pfield
        self.adia_loss_rates_bins.photon_field = pfield
        self.pair_loss_rates_bins.photon_field = pfield
        self.int_rates.photon_field = pfield




from os.path import join

import numpy as np

from prince_cr.util import info, get_AZN
import prince_cr.config as config

from .base import CrossSectionBase


class TabulatedCrossSection(CrossSectionBase):
    """Interface class to read tabulated disintegration cross sections
        Data is expected to be files with names:
        (model_prefix)_egrid.data, (model_prefix)_nonel.data, (model_prefix)_incl_i_j.data 
    
        Data available for Peanut and TALYS with :
        model_prefix = 'PEANUT_IAS' and model_prefix = 'CRP2_TALYS'
        Data available from 1 MeV to 1 GeV
    """

    def __init__(self,
                 model_prefix='PEANUT_IAS',
                 *args,
                 **kwargs):
        self.supports_redistributions = False
        config.max_mass = kwargs.pop('max_mass', config.max_mass)
        CrossSectionBase.__init__(self)
        self._load(model_prefix)
        self._optimize_and_generate_index()

    def _load(self, model_prefix):
        from prince_cr.data import db_handler
        info(2, "Load tabulated cross sections")
        # The energy grid is given in MeV, so we convert to GeV
        photo_nuclear_tables = db_handler.photo_nuclear_db(model_prefix)

        egrid = photo_nuclear_tables["energy_grid"]
        info(2, "Egrid loading finished")

        # Integer idices of mothers and inclusive channels are stored
        # in first column(s)
        pid_nonel = photo_nuclear_tables["inel_mothers"]
        pids_incl = photo_nuclear_tables["mothers_daughters"]

        # the rest of the line denotes the crosssection on the egrid in mbarn,
        # which is converted here to cm^2
        nonel_raw = photo_nuclear_tables["inelastic_cross_sctions"]
        incl_raw = photo_nuclear_tables["fragment_yields"]

        info(2, "Data file loading finished")

        # Now write the raw data into a dict structure
        _nonel_tab = {}
        for pid, csgrid in zip(pid_nonel, nonel_raw):
            if get_AZN(pid)[0] > config.max_mass:
                continue
            _nonel_tab[pid] = csgrid

        # If proton and neutron cross sections are not in contained
        # in the files, set them to 0. Needed for TALYS and CRPropa2
        for pid in [101, 100]:
            if pid not in _nonel_tab:
                _nonel_tab[pid] = np.zeros_like(egrid)

        # mo = mother, da = daughter
        _incl_tab = {}
        for (mo, da), csgrid in zip(pids_incl, incl_raw):
            if get_AZN(mo)[0] > config.max_mass:
                continue
            _incl_tab[mo, da] = csgrid

        self._egrid_tab = egrid
        self._nonel_tab = _nonel_tab
        self._incl_tab = _incl_tab
        # Set initial range to whole egrid
        self.set_range()
        info(2, "Finished initialization")


class CompositeCrossSection(CrossSectionBase):
    """Joins and interpolates between cross section models.
    """

    def __init__(self, model_list):
        """The constructor takes a list of models in the following format::

            $ model_list = [(e_threshold, m_class, m_args),
                            (e_threshold, m_class, m_args)]

        The arguments m1 or m2 are classes derived from
        :class:`CrossSectionBase`. `e_threshold_` is the
        minimal energy for above which a model is used. The maximum
        energy until which a model class is used, is the threshold of the
        next one. m_args are optional arguments passed to the
        constructor of `m_class`.

        Args:
            model_list (list): format as specified above
        """
        CrossSectionBase.__init__(self)
        # References to model instances to be joined
        self.model_refs = None
        self._join_models(model_list)

    def _join_models(self, model_list):

        info(1, "Attempt to join", len(model_list), "models.")

        # Number of modls to join
        nmodels = len(model_list)
        self.model_refs = []
        # Construct instances of models and set ranges where they are valid
        for imo, (e_thr, mclass, margs) in enumerate(model_list):

            # Create instance of a model, passing the provided args
            csm_inst = mclass(*margs)

            if imo < nmodels - 1:
                # If not the highest energy model set both energy limits
                csm_inst.set_range(e_thr, model_list[imo + 1][0])
            else:
                # For the highest energy model, use only minimal energy
                csm_inst.set_range(e_thr)

            # If at least one of the models support redistributions, construct the
            # Interpolator class with redistributions
            if not self.supports_redistributions:
                self.supports_redistributions = csm_inst.supports_redistributions

            # Save reference
            self.model_refs.append(csm_inst)
        #print self.model_refs[1].incl_diff_idcs

        # Create a unique list of nonel cross sections from
        # the combination of all models
        self.nonel_idcs = sorted(
            list(set(sum([m.nonel_idcs for m in self.model_refs], []))))
        # For each ID interpolate the cross sections over entire energy range
        self._nonel_tab = {}
        for mo in self.nonel_idcs:
            self._nonel_tab[mo] = self._join_nonel(mo)

        # Create a unique list of inclusive channels from the combination
        # of all models, no matter if diff or not. The rearrangement
        # is performed in the next steps
        self.incl_idcs_all = sorted(
            list(set(sum([m.incl_idcs for m in self.model_refs], []))))
        self.incl_idcs_all += sorted(
            list(set(sum([m.incl_diff_idcs for m in self.model_refs], []))))

        # Add dynamically generated indices to each model
        newincl = sorted(
            list(
                set(
                    sum([
                        m.generate_incl_channels(self.nonel_idcs)
                        for m in self.model_refs
                    ], []))))
        self.incl_idcs_all = sorted(
            list(set(sum([newincl, self.incl_idcs_all], []))))

        # Collect the channels, that need redistribution functions in a
        # separate list. Put channels that conserve boost into the normal
        # incl_idcs.
        self.incl_diff_idcs = []
        self.incl_idcs = []
        for mother, daughter in self.incl_idcs_all:
            if self.is_differential(mother, daughter):
                info(10, 'Daughter has redistribution function', mother,
                     daughter)
                self.incl_diff_idcs.append((mother, daughter))
            else:
                info(10, 'Mother and daughter conserve boost', mother,
                     daughter)
                self.incl_idcs.append((mother, daughter))

        # Join the Egrid
        self._egrid_tab = np.concatenate([m.egrid for m in self.model_refs])
        self.set_range()

        # Join the boost conserving channels
        self._incl_tab = {}
        for mother, daughter in self.incl_idcs:
            self._incl_tab[(mother, daughter)] = self._join_incl(
                mother, daughter)

        # Join the redistribution channels
        self._incl_diff_tab = {}
        for mother, daughter in self.incl_diff_idcs:
            self._incl_diff_tab[(mother, daughter)] = self._join_incl_diff(
                mother, daughter)

        self._update_indices()
        self._optimize_and_generate_index()

    def _join_nonel(self, mother):
        """Returns the non-elastic cross section of the joined models.
        """

        info(5, 'Joining nonelastic cross sections for', mother)

        egrid = []
        nonel = []
        for model in self.model_refs:
            e, csec = model.nonel(mother)
            egrid.append(e)
            nonel.append(csec)

        #return np.concatenate(nonel)
        return np.concatenate(egrid), np.concatenate(nonel)

    def _join_incl(self, mother, daughter):
        """Returns joined incl cross sections."""

        info(5, 'Joining inclusive cross sections for channel',
             (mother, daughter))
        egrid = []
        incl = []

        for model in self.model_refs:
            e, csec = model.incl(mother, daughter)
            egrid.append(e)
            incl.append(csec)
        #print np.concatenate(egrid), np.concatenate(incl)
        #print '---'*30
        #return np.concatenate(incl)
        return np.concatenate(egrid), np.concatenate(incl)

    def _join_incl_diff(self, mother, daughter):
        """Returns joined incl diff cross sections.

        The function assumes the same `x` bins for all models.
        """

        info(5, 'Joining inclusive differential cross sections for channel',
             (mother, daughter))

        egrid = []
        incl_diff = []

        # Get an x grid from a model which supports it
        for model in self.model_refs:
            if model.supports_redistributions:
                self.xbins = model.xbins
                break
        if self.xbins is None:
            raise Exception('Redistributions requested but none of the ' +
                            'models supports it')

        for model in self.model_refs:
            egr, csec = None, None
            if config.debug_level > 1:
                if not np.allclose(self.xbins, model.xbins):
                    raise Exception('Unequal x bins. Aborting...',
                                    self.xbins.shape, model.xbins)
            if (mother, daughter) in model.incl_diff_idcs:
                egr, csec = model.incl_diff(mother, daughter)
                info(10, model.mname, mother, daughter, 'is differential.')

            elif (mother, daughter) in model.incl_idcs:
                # try to use incl and extend by zeros
                egr, csec_1d = model.incl(mother, daughter)
                print(mother, daughter, csec_1d.shape)
                # no x-distribution given, so x = 1
                csec = self._arange_on_xgrid(csec_1d)
                info(1, model.mname, mother, daughter,
                     'not differential, x=1.')
            else:
                info(
                    5, 'Model', model.mname, 'does not provide cross',
                    'sections for channel {0}/{1}. Setting to zero.'.format(
                        mother, daughter))
                # Tried with reduced energy grids to save memory, but
                # matrix addition in decay chains becomes untrasparent
                # egr = np.array((model.egrid[0], model.egrid[-1]))
                # csec = np.zeros((len(self.xbins) - 1, 2))
                egr = model.egrid
                csec = np.zeros((len(self.xbins) - 1, model.egrid.size))

            egrid.append(egr)
            incl_diff.append(csec)

        #return np.concatenate(incl_diff, axis=1)
        return np.concatenate(egrid), np.concatenate(incl_diff, axis=1)



import numpy as np

from prince_cr.util import get_2Dinterp_object, get_interp_object, info, get_AZN

from .base import CrossSectionBase


class ResponseFunction(object):
    """Redistribution Function based on Crossection model

        The response function is the angular average crosssection
    """
    def __init__(self, cross_section):
        self.cross_section = cross_section

        self.xcenters = cross_section.xcenters

        # Copy indices from CrossSection Model
        self.nonel_idcs = cross_section.nonel_idcs
        self.incl_idcs = cross_section.incl_idcs
        self.incl_diff_idcs = cross_section.incl_diff_idcs

        # Dictionary of reponse function interpolators
        self.nonel_intp = {}
        self.incl_intp = {}
        self.incl_diff_intp = {}
        self.incl_diff_intp_integral = {}
        
        self._precompute_interpolators()

    # forward is_differential() to CrossSectionBase
    # that might break in the future...
    def is_differential(self, mother, daughter):
        return CrossSectionBase.is_differential(self, mother, daughter)

    def get_full(self, mother, daughter, ygrid, xgrid=None):
        """Return the full response function :math:`f(y) + g(y) + h(x,y)`
        on the grid that is provided. xgrid is ignored if `h(x,y)` not in the channel.
        """
        if xgrid is not None and ygrid.shape != xgrid.shape:
            raise Exception('ygrid and xgrid do not have the same shape!!')
        if get_AZN(mother)[0] < get_AZN(daughter)[0]:
            info(
                3,
                'WARNING: channel {:} -> {:} with daughter heavier than mother!'
                .format(mother, daughter))

        res = np.zeros(ygrid.shape)

        if (mother, daughter) in self.incl_intp:
            res += self.incl_intp[(mother, daughter)](ygrid)
        elif (mother, daughter) in self.incl_diff_intp:
            #incl_diff_res = self.incl_diff_intp[(mother, daughter)](
            #    xgrid, ygrid, grid=False)
            #if mother == 101:
            #    incl_diff_res = np.where(xgrid < 0.9, incl_diff_res, 0.)
            #res += incl_diff_res
            #if not(mother == daughter):
            res += self.incl_diff_intp[(mother, daughter)].inteval(xgrid,
                                                                   ygrid,
                                                                   grid=False)

        if mother == daughter and mother in self.nonel_intp:
            # nonel cross section leads to absorption, therefore the minus
            if xgrid is None:
                res -= self.nonel_intp[mother](ygrid)
            else:
                diagonal = xgrid == 1.
                res[diagonal] -= self.nonel_intp[mother](ygrid[diagonal])

        return res

    def get_channel(self, mother, daughter=None):
        """Reponse function :math:`f(y)` or :math:`g(y)` as
        defined in the note.

        Returns :math:`f(y)` or :math:`g(y)` if a daughter
        index is provided. If the inclusive channel has a redistribution,
        :math:`h(x,y)` will be returned

        Args:
            mother (int): mother nucleus(on)
            daughter (int, optional): daughter nucleus(on)

        Returns:
            (numpy.array) Reponse function on self._ygrid_tab
        """
        from scipy import integrate

        cs_model = self.cross_section
        egrid, cross_section = None, None

        if daughter is not None:
            if (mother, daughter) in self.incl_diff_idcs:
                egrid, cross_section = cs_model.incl_diff(mother, daughter)
            elif (mother, daughter) in self.incl_idcs:
                egrid, cross_section = cs_model.incl(mother, daughter)
            else:
                raise Exception(
                    'Unknown inclusive channel {:} -> {:} for this model'.
                    format(mother, daughter))
        else:
            egrid, cross_section = cs_model.nonel(mother)

    # note that cumtrapz works also for 2d-arrays and will integrate along axis = 1
        integral = integrate.cumtrapz(egrid * cross_section, x=egrid)
        ygrid = egrid[1:] / 2.

        return ygrid, integral / (2 * ygrid**2)

    def get_channel_scale(self, mother, daughter=None, scale='A'):
        """Returns the reponse function scaled by `scale`.

        Convenience funtion for plotting, where it is important to
        compare the cross section/response function per nucleon.

        Args:
            mother (int): Mother nucleus(on)
            scale (float): If `A` then nonel/A is returned, otherwise
                           scale can be any float.

        Returns:
            (numpy.array, numpy.array): Tuple of Energy grid in GeV,
                                        scale * inclusive cross section
                                        in :math:`cm^{-2}`
        """

        ygr, cs = self.get_channel(mother, daughter)

        if scale == 'A':
            scale = 1. / get_AZN(mother)[0]

        return ygr, scale * cs

    def _precompute_interpolators(self):
        """Interpolate each response function and store interpolators.

        Uses :func:`prince.util.get_interp_object` as interpolator.
        This might result in too many knots and can be subject to
        future optimization.
        """

        info(2, 'Computing interpolators for response functions')

        info(5, 'Nonelastic response functions f(y)')
        self.nonel_intp = {}
        for mother in self.nonel_idcs:
            self.nonel_intp[mother] = get_interp_object(
                *self.get_channel(mother))

        info(5, 'Inclusive (boost conserving) response functions g(y)')
        self.incl_intp = {}
        for mother, daughter in self.incl_idcs:
            self.incl_intp[(mother, daughter)] = get_interp_object(
                *self.get_channel(mother, daughter))

        info(5, 'Inclusive (redistributed) response functions h(y)')
        self.incl_diff_intp = {}
        for mother, daughter in self.incl_diff_idcs:
            ygr, rfunc = self.get_channel(mother, daughter)
            self.incl_diff_intp[(mother, daughter)] = get_2Dinterp_object(
                self.xcenters, ygr, rfunc, self.cross_section.xbins)

            from scipy.integrate import cumtrapz
            integral = cumtrapz(rfunc, ygr, axis=1,initial=0)
            integral = cumtrapz(integral, self.xcenters, axis=0,initial=0)

            self.incl_diff_intp_integral[(mother, daughter)] = get_2Dinterp_object(
                self.xcenters, ygr, integral, self.cross_section.xbins)


"""Module implementing relations from paper

Functions implementing formulas from paper 
http://stacks.iop.org/1402-4896/49/i=3/a=004
and others, to construct the inclusive cross
sections.
"""
import itertools
import sys
from collections import Counter
from os.path import join
from pickle import load

from numpy import array, exp, inf, linspace, sum

from prince_cr.util import get_AZN
from prince_cr.data import spec_data
from prince_cr.config import config

# JH: This file gave some linter errors for me, disabled for now
# pylint: skip-file

# class mydict(dict):
#     def __getitem__(self, key):
#         if key not in self:
#             self.__setitem__(key, 0)
            
#         return dict.__getitem__(self, key)

# listing functions to create species tables
def list_species_by_mass(Amax, tau=inf):
    '''Returns a dictionary with the species stable enough
    to be produced in spallation.
    '''
    species = {}
    for nuc in sorted([k for k in spec_data.keys() if isinstance(k, int)]):
        if (nuc < 100) or (spec_data[nuc]['lifetime'] < tau):
            continue
        At, _, _ = get_AZN(nuc)

        if At in species:
            species[At].append(nuc)
        else:
            species[At] = [nuc]

    return species


def partitions(n):
    """Returns all integer combinations that add up to n
    and no summand is higher than 4
    """	
    # base case of recursion: zero is the sum of the empty list
    if n == 0:
        yield []
        return
        
    # modify partitions of n-1 to form partitions of n
    for p in partitions(n-1):
        yield [1] + p
        if p and (len(p) < 2 or p[1] > p[0]):
            if p[0] < 4:
                yield [p[0] + 1] + p[1:]


def combinations(x, y):
    '''Returns possible combinations of nuclei with mass A<=4
    such that they contain x protons and y neutrons.
    '''
    ncos_id = int((x + y)*100 + x)
    mass_partitions = partitions(x + y)

    for mass_partition in mass_partitions:
        mass_partition = [f for f in mass_partition if f > 1]
        species = [species_by_mass[Af] for Af in mass_partition if Af in species_by_mass]
        for c in list(itertools.product(*species)):
            _, z, n = get_AZN(sum(c))
            if (z <= x) and (n <= y):
                yield int(y-n)*(100, ) + int(x-z)*(101, ) + c


def residual_multiplicities():
    '''Makes a dictionary where the ncos id with x,y number of protons and 
    neutrons emitted are the keys, and the multiplicities of species between A=1-4
    are given, normalized such that they add up to x protons and y neutrons.
    This is a function to precompute the table which is used to find the multiplicities
    of the empirical photomeson model.
    '''
    spalled_nucleons = []
    for Am in species_by_mass:
        for mother in species_by_mass[Am]:
            for A_big_frag in range(Am/2, Am-1):
                for big_frag in species_by_mass[A_big_frag]:
                    if big_frag % 100 > mother % 100:
                        continue
                    spalled_frag = mother - big_frag
                    spalled_nucleons.append(spalled_frag)
                    # print spalled_nucleons[-1]
    residual_list = {}
    count = 0
    last = 0
    print('Completed.... ', 0)
    cant = float(len(set(spalled_nucleons)))
    for tot in set(spalled_nucleons):
        count+=1
        # print '--', tot, '--', '{:3.3f}'.format(count/cant)
        if int(100 * count / cant) >= last + 5:
            print('Completed.... ', int(100 * count / cant))
            last += 5
        _, x, y = get_AZN(tot)
        counts = Counter([e for elem in combinations(x, y) for e in elem])
        suma = 0
        for k, v in counts.items():
            suma += k * v
        for k in counts:
            counts[k] *= tot/float(suma)
        residual_list[tot] = counts.copy()

    return residual_list

# local lookup tables for efficiency
# species_by_mass = list_species_by_mass(56, config['tau_dec_threshold'])
species_by_mass = list_species_by_mass(56, 0.)
# resmul = residual_multiplicities()
small_frags_relative_yields_filename = join(config.data_dir, 'small_frags_relative_yields.pkl')
with open(small_frags_relative_yields_filename, 'rb') as f:
    # it's faster to pickle.load a precomputed resmul
    resmul = load(f,encoding='latin1')


#### empirical relations from reference...
def cs_gpi(A):
    """Cross section for pion photoproduction averaged over 
    E[.14, 1.] GeV
    
    Average cross section values obtained by Monte Carlo
    simulations, for nuclei of different masses, starting at
    7Li. It includes the production of pions of all types.
    
    Arguments:
        A {int} -- Number of nucleons in the target nucleus
    
    Returns:
        float -- Mean cross section in miliibarn
    """
    return 0.027 * A**.847


def cs_gn(A):
    """Cross section for A(g,n)X averaged over E[.3, 1.] GeV
    
    Returns cross section of photoneutron production averaged
    over the energy range [.3, 1.] GeV, in milibarn units.
    
    Arguments:
        A {int} -- Nucleon number of the target nucleus
    """

    return 0.104 * A**0.81


def xm(A):
    """Returns the maximum number of emmited neutrons in a 
    A(g,xn)X reaction
    
    See function cs_gxn()
    
    Arguments:
        A {int} -- 
    
    Returns:
        int -- Maximum number of neutrons to be produced
    """
    return int(1.4 * A**.457)


def cs_gxn(A, x=2):
    """Cross section for A(g,xn)X averaged over E[.2, 1.] GeV
    
    Returns cross section of photoneutrons production (x > 1)
    averaged over the energy range [.3, 1.] GeV, in milibarn 
    units. The number of neutrons x is checked to be lower than
    xm (see function definition xm).
    
    Arguments:
        A {int} -- Nucleon number of the target nucleus
        x {int} -- Number of neutrons produced (1 < x < xm)
    """

    if 1 < x < xm(A):
        k = 37. * A ** -.924
        return 0.187 * A**0.684 * exp(-k * (x - 1)**1.25)
    else:
        return 0


def cs_gxn_all(A):
    """Cross section for A(g,xn)X averaged over E[.2, 1.] GeV
    
    Returns cross section of photoneutrons production (x > 1)
    averaged over the energy range [.3, 1.] GeV, in milibarn 
    units. The number of neutrons x is checked to be lower than
    xm (see function definition xm).
    
    Arguments:
        A {int} -- Nucleon number of the target nucleus
        x {int} -- Number of neutrons produced (1 < x < xm)
    """
    cs_gxn_summed = 0
    for xi in range(2, xm(A)):
        cs_gxn_summed += cs_gxn(A, xi)

    return cs_gxn_summed


def cs_gp(Z=1, **kwargs):
    """Cross section for A(g,p)X averaged over E[.3, .8] GeV
    
    Returns cross section of photoproduction of a proton averaged
    over the energy range [.3, .8] GeV, in milibarn units.
    
    Arguments:
        Z {int} -- Number of protons of the target nucleus
        A {int} -- (optional) 
    """
    if 'A' in kwargs:
        return 0.078 * kwargs['A']**0.5
    else:
        return 0.115 * Z**0.5


def cs_gSp(Z, A, x=1, y=1):
    """Cross section for spallation averaged over E[.2, 1.] GeV
    
    Returns cross section of photoproduction of multiple
    protons and neutrons in a spallation procees, averaged
    over the energy range [.2, 1.] GeV, in milibarn units.
    
    Arguments:
        Z {int} -- Number of protons of the target nucleus
        A {int} -- Number of nucleons of the target nucleus
                   A <= 90
        x {int} -- Number of protons produced. Default is 1.
                   1 <= x <= Z/2
        y {int} -- Number of neutron produced. Default is 1.
                   1 <= y <= (A - Z)/2

    """
    K = 0.466
    a = float(Z) / (A - Z)
    C = 2.3 * a - 1.044
    E = 446. / A
    if E < 21.:
        cs_M = 15.7 / E**1.356
    elif 21. < E:
        cs_M = 0.248
    if E < 10.:
        B = 3.03 / E**1.06
    elif E > 10.:
        B = 0.25

    return cs_M * exp(-B * (x - 1) - K*(x - C*a*y)**2)


def cs_gSp_all(Z, A):
    """Cross section summed for all possible spallation events
    """	
    mother = 100*A + Z
    cs_tot = 0
    for A_big_frag in range(A/2, A-1):
        for big_frag in species_by_mass[A_big_frag]:
            _, x, y = get_AZN(mother - big_frag)
            spalled_id = 100*(x+y) + x
            
            if (x < 1) or (y < 1):
                # in spallation at least a neutron and proton escape
                continue

            cs_frag = cs_gSp(Z, A, x, y)
            cs_tot += cs_frag

    return cs_tot


def cs_gSp_all_inA(A):
    """Cross section summed for all possible spallation events
    """
    n = 0
    cs_summed = 0
    cs_vals = []
    if A in species_by_mass:
        for nuc in species_by_mass[A]:
            A, Z, N = get_AZN(nuc)        
            cs_summed = cs_gSp_all(Z, A)
            cs_vals.append(cs_gSp_all(Z, A))
            n += 1
    
    # return cs_summed / n
    return max(cs_vals)


def cs_tot(A):
    '''Determines the norm for the empirical formulas
    such that the addition of inclusive cross sections
    does not fluctuate with A. Renormalization value 
    depends on the mass and is set to the mean of the
    total empirical cross section per nucleon over 
    a range of A in A=4-55.
    '''
    csp = cs_gp(A=A)
    cspi = cs_gpi(A)
    csn = cs_gn(A)
    csxn = cs_gxn_all(A)
    csSpal = cs_gSp_all_inA(A)
    
    cstot = csp + cspi + csn + csxn + csSpal

    return cstot * cs_norm[A]


#### inclusive cross sections derived from relations above, and related functions

def partition_probability(partition, A, beta=.1):
    """Gives partitions probabilities as exp(-beta*Ei)
    where Ei is the sum of binding energies of the components of the partition.
    
    Given a partition (list of mass fragments {A_j}), the sets of all possible
    species per mass {S^A_k} are taken. Then all possible combinations corresponding
    to the partition are built C_l={S^j_m} and the combination is given a probability
    based on it's energy: P_l = exp(-beta*E_l). The probablities of all combinations
    are normalized to unity (P_l = P_l / sum(P_l))
    
    Arguments:
        partition {[list]} -- list of [A_k] masses into which the original nucleus was split

    Returns:
        combinations -- a list of possible nuclei that make up the A partition given
        yields -- the probabilities of each of the partitions, obtained by a statistical
            argument (i.e. canonical ensemble )
    """

    species = [species_by_mass[Af] for Af in partition if Af in species_by_mass]
    combinations = list(itertools.product(*species))

    yields = []
    for combination in combinations:
        Ei = A
        for nuc in combination:
            Ei -= spec_data[nuc]['mass']
        
        yields.append(exp(-beta*Ei))

    tot_yields = sum(yields)
    yields = [y/tot_yields for y in yields]
    
    return combinations, yields


def partition_probability(combinations, A, beta=.1):
    """Gives partitions probabilities as exp(-beta*Ei)
    where Ei is the sum of binding energies of the components of the partition.
    
    Given a partition (list of mass fragments {A_j}), the sets of all possible
    species per mass {S^A_k} are taken. Then all possible combinations corresponding
    to the partition are built C_l={S^j_m} and the combination is given a probability
    based on it's energy: P_l = exp(-beta*E_l). The probablities of all combinations
    are normalized to unity (P_l = P_l / sum(P_l))
    
    Arguments:
        partition {[list]} -- list of [A_k] masses into which the original nucleus was split

    Returns:
        combinations -- a list of possible nuclei that make up the A partition given
        yields -- the probabilities of each of the partitions, obtained by a statistical
            argument (i.e. canonical ensemble )
    """

    yields = []
    for combination in combinations:
        Ei = A
        for nuc in combination:
            Ei -= spec_data[nuc]['mass']
        
        yields.append(exp(-beta*Ei))

    tot_yields = sum(yields)
    yields = [y/tot_yields for y in yields]
    
    return combinations, yields


def gxn_multiplicities(mother):
    """Multiplicities for multineutron emission A(g,xn)X
    
    Arguments:
        A {int} -- Nucleon number of the target nucleus
    """
    cs_sum = 0
    cs_gxn_incl = {100:0}
    Am, _, _ = get_AZN(mother)

    for xi in range(2, xm(Am)):
        cs = cs_gxn(Am, xi)
        cs_gxn_incl[100] += xi*cs
        cs_gxn_incl[mother - xi*100] = cs

        cs_sum += cs

    if cs_sum == 0:
        cs_sum = inf

    for dau in cs_gxn_incl:
        if cs_gxn_incl[dau] != 0:
            cs_gxn_incl[dau] /= cs_sum
    
    return cs_gxn_incl


def spallation_multiplicities(mother):
    '''Calculates the inclusive cross sections of all fragments
    of mothter species (moter is a neucos id) for spallation
    '''
    Am, Zm, _ = get_AZN(mother)
    
    incl_tab = {}
    cs_sum = 0
    for A_big_frag in range(Am//2, Am-1):
        for big_frag in species_by_mass[A_big_frag]:
            _, x, y = get_AZN(mother - big_frag)
            spalled_id = 100*(x+y) + x
            
            if (x < 1) or (y < 1):
                # in spallation at least a neutron and proton escape
                continue
            
            cs_frag = cs_gSp(Zm, Am, x, y)
            cs_sum += cs_frag  # sum of all cross sections to normalize incl_tab
            
            if big_frag in incl_tab:
                incl_tab[big_frag] += cs_frag
            else:
                incl_tab[big_frag] = cs_frag

            # get low fragment incl_tab from using Counter on a prepared list with x, y outputs
            for dau in resmul[spalled_id]:
                if dau in incl_tab:
                    incl_tab[dau] += cs_frag * resmul[spalled_id][dau]
                else:
                    incl_tab[dau] = cs_frag * resmul[spalled_id][dau]
    for dau in incl_tab:
        incl_tab[dau] /= cs_sum  # all spallation cross section should match total spallation cross section

    return incl_tab


def cs_Rincl(Z, A, yields):
    """Returns incl of residual production
    
    [description]
    
    Arguments:
        Z {[type]} -- [description]
        A {[type]} -- [description]
    """
    Amax = max(yields.keys())
    nuclist = [100, 101]
    csilist = [cs_nincl(Z, A), cs_pincl(Z, A)]
    sub_frags = {}
    for nz in range(0, int(Z / 2.) + 1):
        for nn in range(0, int((A - Z) / 2.) + 1):
            Ared = min(Amax, nn + nz)

            # print nn, nz, A-Z, Z
            if nn + nz == 0:
                # add pion contribution
                nuclist += [A*100 + Z,
                            A*100 + Z - 1,
                            A*100 + Z + 1]
                csilist += [cs_gpi(A) / 3,] * 3
                continue

            cs_incl = 0
            new_frag = 0
            if (nn == 1) and (nz == 0):
                cs_incl = cs_gn(A)
            elif (nn > 1) and (nz == 0):
                cs_incl = cs_gxn(A, nn)
            elif (nn == 0) and (nz == 1):
                cs_incl = cs_gp(Z)
            elif (nn >= 1) and (nz >= 1):
                cs_incl = cs_gSp(Z, A, nz, nn)
                new_frag = 100 * Ared + Ared/2 - nz
                if new_frag > 0:
                    csilist.append(cs_incl)
                    nuclist.append(new_frag)				
                sub_frags = yields[Ared]

            if cs_incl > 0:
                csilist.append(cs_incl)
                nuclist.append(100 * (A - nn - nz) + Z - nz)
                if sub_frags:
                    suma = 0
                    for nuc, val in sub_frags.items():
                        Af, _, _ = get_AZN(nuc)
                        suma += Af * val
                    norm = Ared / suma

                    for nuc, val in sub_frags.items():
                        if nuc in nuclist:
                            csilist[nuclist.index(nuc)] += val * norm * cs_incl
                        else:
                            csilist.append(val * norm * cs_incl)
                
    return nuclist, csilist


def multiplicity_table(mother):
    '''Returns a dict with the multiplicities
    for all fragments from mother is contained.
    The differentence with spallation_inclusive is that
    here all processes are contained
    '''
    gxn_mult = gxn_multiplicities(mother)
    sp_mult = spallation_multiplicities(mother)
    
    Am, Zm, _ = get_AZN(mother)

    cspi = cs_gpi(Am)
    csp = cs_gp(A=Am)
    csn = cs_gn(Am)
    csxn = cs_gxn_all(Am)
    cs_tot = .28 * Am
    csSp = cs_tot - (cspi + csp + csn + csxn)

    multiplicities = {100: 1.*csn/cs_tot,
                      101: 1.*csp/cs_tot,
                      mother - 100: 1.*csn/cs_tot,
                      mother - 101: 1.*csp/cs_tot,}

    for dau, mult in gxn_mult.items():
        if dau in multiplicities:
            multiplicities[dau] += mult * csxn / cs_tot
        else:
            multiplicities[dau] = mult * csxn / cs_tot

    for dau, mult in sp_mult.items():
        if dau in multiplicities:
            multiplicities[dau] += mult * csSp / cs_tot
        else:
            multiplicities[dau] = mult * csSp / cs_tot

    return multiplicities


def main():
    pass
    # resmul = residual_multiplicities()
    print(list(resmul.keys()))

if __name__ == '__main__':
    main()



"""Defines source models for cosmic ray propagation

    The standard interface requires to UHECRSolvers requires
    that each source defines a methods injection_rate(self, z)
"""

from abc import ABCMeta, abstractmethod

import numpy as np

from prince_cr.util import info


class CosmicRaySource(object, metaclass=ABCMeta):
    def __init__(self,
                 prince_run,
                 ncoids=None,
                 params=None,
                 norm=1.,
                 m='flat',
                 *args,
                 **kwargs):
        # read out standard information from core class
        self.cr_grid = prince_run.cr_grid.grid
        self.dim_states = prince_run.dim_states
        self.norm = norm
        self.spec_man = prince_run.spec_man

        # read out the input parameters, this is a dictionary with particle ids as key
        # the parameters are interpreted by each child class.
        self.params = params
        self.ncoids = np.array(ncoids if ncoids is not None else list(params.keys()))
        self.ncoids.sort()
        self.source_evo_m = m
        self._compute_injection_grid()

    def _compute_injection_grid(self):
        """Precompute the injection for all species on a single grid.
        
        Assumes that the injection is factorized in E and z"""
        self.injection_grid = np.zeros(self.dim_states)

        for pid in self.ncoids:
            if pid in self.params:
                params = self.params[pid]
            else:
                params = params

            info(
                4, 'Injecting particle {:} with parameters {:}'.format(
                    pid, params))
            inj_spec = self.spec_man.ncoid2sref[pid]
            self.injection_grid[inj_spec.sl] = self.injection_spectrum(
                pid, self.cr_grid, params)

    def integrated_lum(self, Emin=1e9):
        """Integrates the injected spectrum for all particles
        
        Args:
            Emin (float): lower limit of the integration
    
        Returns:
            float array: tuple with integrated number and luminosity for each species 
        """
        num_int = np.zeros_like(self.ncoids, dtype=np.float)
        lum_int = np.zeros_like(self.ncoids, dtype=np.float)

        from scipy.integrate import trapz
        for idx, pid in enumerate(self.ncoids):
            # get the inection for the species and subsitute back from E_A =  E / A to E
            s = self.spec_man.ncoid2sref[pid]
            A = s.A
            egrid = self.cr_grid * A
            injec = self.injection_grid[s.sl] / A

            # Get the ranges where the energy is greater than Emin and integrate
            mask = np.argwhere(egrid > Emin)
            mask = mask.flatten()
            num_int[idx] = trapz(injec[mask], egrid[mask])
            lum_int[idx] = trapz(injec[mask] * egrid[mask], egrid[mask])
        return num_int, lum_int

    def injection_rate(self, z):
        """Returns the injection rate on the given self.cr_grid

        Args:
            z (float): redshift

        Returns:
            float array: array of the same length as self.cr_grid
        """
        return self.norm * self.evolution(z) * self.injection_grid

    def injection_rate_single(self, pid, energy, z):
        """Return the injection rate for a single energy and redshift

        Args:
            pid (int): particle id for which to return the injection
            energy (float): single float or array for which to return the injection
            z (float): redshift

        Return:
            float: injection on the grid defined by energy
        """
        if pid in self.params:
            params = self.params[pid]
        return self.norm * self.evolution(z) * self.injection_spectrum(
            pid, energy, params)

    @abstractmethod
    def injection_spectrum(self, pid, energy, params):
        """Prototype to be defined in each child class"""

    def evolution(self, z):
        """Returns the source evolution function at given redshift

        Note: The source evolution will depend on self.source_evo_m
              This can be a float, a keyword or a tuple of both.
              See code for supported combinations

        Args:
            z (float): redshift

        Return:
            float: Relative source evolution, typically normalized to 1 at z=0
        """
        from .cosmology import star_formation_rate, grb_rate_wp, agn_rate

        # Check if negative z is called
        if z < 0:
            raise Exception(
                'Source evolution not defined for negative z = {:}'.format(z))

        # flat source evolution
        if self.source_evo_m == 'flat':
            return 1.
        # simple source evolution (1+z)**m
        elif type(self.source_evo_m) is float:
            return (1 + z)**self.source_evo_m * star_formation_rate(z)
        # for a tuple decide based on keyword
        elif type(self.source_evo_m) is tuple:
            mkwd, mval = self.source_evo_m
            # in these cases the evolution is taken from keyword and scaled by m
            if mkwd == 'SFR':
                return (1 + z)**mval * star_formation_rate(z)
            elif mkwd == 'GRB':
                return (1 + z)**mval * grb_rate_wp(z)
            elif mkwd == 'AGN':
                return (1 + z)**mval * agn_rate(z)
            elif mkwd == 'TDE':
                return (1 + z)**(mval-3.)
            elif mkwd == 'simple':
                return (1 + z)**mval
            # local evolution as (1+z)**m and flat beyond z = 1
            elif mkwd == 'simple_flat':
                if z <= 1:
                    return (1 + z)**mval
                else:
                    return (1 + 1)**mval
            # local evolution as (1+z)**m and a break of n = m - 3.6 at z = 1
            elif mkwd == 'simple_SFR':
                if z <= 1:
                    return (1 + z)**mval
                else:
                    return (1 + 1)**3.6 * (1 + z)**(mval - 3.6)
        else:
            raise Exception('Unknown source evo type: {:}'.format(
                self.source_evo_m))


class SimpleSource(CosmicRaySource):
    """Simple source class with spectral index and cutoff

        inj(E) = norm * E**-gamma * exp(- E / Emax)

        params defined as {pid: gamma, Emax, norm}
    """

    def injection_spectrum(self, pid, energy, params):
        gamma, emax, norm = params
        inj_spec = self.spec_man.ncoid2sref[pid]
        emax = emax / inj_spec.A
        result = norm * energy**(-gamma) * np.exp(-energy / emax)

        return result


class RigdityCutoffSource(CosmicRaySource):
    """Simple source class with spectral index and rigidity dependent cutoff

        inj(E) = norm * E**-gamma * exp(- E / Z * Rcut)

        params defined as {pid: gamma, Rcut, norm}
    """

    def injection_spectrum(self, pid, energy, params):
        spectral_index, rcut, relnorm = params
        inj_spec = self.spec_man.ncoid2sref[pid]
        A = float(inj_spec.A)
        emax = rcut * inj_spec.Z
        e_k = A * energy
        result = relnorm * A * (e_k / 1e9)**(
            -spectral_index) * np.exp(1 - e_k / emax)

        return result


class AugerFitSource(CosmicRaySource):
    """Simple source class with spectral index and rigidity dependent cutoff
        Defined to be parrallel for all species below the cutoff as in Auger Combined Fit paper

        if E <  Z * Rcut:
            inj(E) = norm * E**-gamma
        else:
            inj(E) = norm * E**-gamma * exp(- E / Z * Rcut)

        params defined as {pid: gamma, Rcut, norm}
    """

    def injection_spectrum(self, pid, energy, params):
        spectral_index, rcut, relnorm = params
        inj_spec = self.spec_man.ncoid2sref[pid]
        A = float(inj_spec.A)
        emax = rcut * inj_spec.Z
        e_k = A * energy
        result = relnorm * A * (e_k / 1e9)**(-spectral_index) * np.where(
            e_k < emax, 1., np.exp(1 - e_k / emax))

        return result


class RigidityFlexSource(CosmicRaySource):
    """Simple source class with spectral index and rigidity dependent cutoff
        Parameter alpha to scaled the rigidity dependence

        inj(E) = norm * E**-gamma * exp(- E / Z**alpha * Rcut)

        params defined as {pid: gamma, Rcut, alpha, norm}
    """

    def injection_spectrum(self, pid, energy, params):
        spectral_index, rcut, alpha, relnorm = params
        inj_spec = self.spec_man.ncoid2sref[pid]
        A = float(inj_spec.A)
        emax = rcut * inj_spec.Z**alpha
        e_k = A * energy
        result = relnorm * A * (e_k / 1e9)**(-spectral_index) * np.where(
            e_k < emax, 1., np.exp(1 - e_k / emax))

        return result


class SpectrumSource(CosmicRaySource):
    """Source class with the spectrum defined externally by an array
        The spectrum might be interpolated as needed
       
        params defined as {pid: egrid, specgrid}
    """

    def injection_spectrum(self, pid, energy, params):
        egrid, specgrid = params
        inj_spec = self.spec_man.ncoid2sref[pid]
        egrid = egrid / inj_spec.A
        specgrid = specgrid * inj_spec.A
        result = np.interp(energy, egrid, specgrid, left=0., right=0.)

        return result



# Test whether cross section are correctly created.

import numpy as np

import prince_cr.config as config
from prince_cr import cross_sections, photonfields, core
from prince_cr.solvers import UHECRPropagationSolverBDF
from prince_cr.cr_sources import AugerFitSource

config.x_cut = 1e-4
config.x_cut_proton = 1e-2
config.tau_dec_threshold = np.inf
config.max_mass = 14
config.debug_level = 0  # suppress info statements

pf = photonfields.CombinedPhotonField(
    [photonfields.CMBPhotonSpectrum, photonfields.CIBGilmore2D])
cs = cross_sections.CompositeCrossSection([
    (0., cross_sections.TabulatedCrossSection, ('CRP2_TALYS', )),
    (0.14, cross_sections.SophiaSuperposition, ())
])

prince_run_talys = core.PriNCeRun(max_mass=14,
                                  photon_field=pf,
                                  cross_sections=cs)

import unittest


class TestCsec(unittest.TestCase):
    def test_propagation(self):
        solver = UHECRPropagationSolverBDF(initial_z=1.,
                                           final_z=0.,
                                           prince_run=prince_run_talys,
                                           enable_pairprod_losses=True,
                                           enable_adiabatic_losses=True,
                                           enable_injection_jacobian=True,
                                           enable_partial_diff_jacobian=True)

        solver.add_source_class(
            AugerFitSource(prince_run_talys,
                           norm=1e-50,
                           params={
                               101: (0.96, 10**9.68, 20.),
                               402: (0.96, 10**9.68, 50.),
                               1407: (0.96, 10**9.68, 30.)
                           }))
        solver.solve(dz=1e-3, verbose=False, progressbar=False)

        self.assertEqual(solver.known_species, prince_run_talys.spec_man.known_species)

if __name__ == '__main__':
    unittest.main()


"""Contains solvers, i.e. integrators, kernels, steppers, for PriNCe."""

import numpy as np

from prince_cr.cosmology import H
from prince_cr.data import PRINCE_UNITS, EnergyGrid
from prince_cr.util import info
import prince_cr.config as config

from .partial_diff import DifferentialOperator, SemiLagrangianSolver

class UHECRPropagationResult(object):
    """Reduced version of solver class, that only holds the result vector and defined add and multiply
    """

    def __init__(self, state, egrid, spec_man):
        self.spec_man = spec_man
        self.egrid = egrid
        self.state = state

    def to_dict(self):
        dic = {}
        dic['egrid'] = self.egrid
        dic['state'] = self.state
        dic['known_spec'] = self.known_species

        return dic

    @classmethod
    def from_dict(cls, dic):
        egrid = dic['egrid']
        edim = egrid.size
        state = dic['state']
        known_spec = dic['known_spec']

        from ..data import SpeciesManager
        spec_man = SpeciesManager(known_spec, edim)

        return cls(state, egrid, spec_man)

    @property
    def known_species(self):
        return self.spec_man.known_species

    def __add__(self, other):
        cumstate = self.state + other.state

        if not np.array_equal(self.egrid, other.egrid):
            raise Exception(
                'Cannot add Propagation Results, they are defined in different energy grids!'
            )
        if not np.array_equal(self.known_species, other.known_species):
            raise Exception(
                'Cannot add Propagation Results, have different species managers!'
            )
        else:
            return UHECRPropagationResult(cumstate, self.egrid, self.spec_man)

    def __mul__(self, number):
        if not np.isscalar(number):
            raise Exception(
                'Can only multiply result by scalar number, got type {:} instead!'.
                format(type(number)))
        else:
            newstate = self.state * number
            return UHECRPropagationResult(newstate, self.egrid, self.spec_man)

    def get_solution(self, nco_id):
        """Returns the spectrum in energy per nucleon"""
        spec = self.spec_man.ncoid2sref[nco_id]
        return self.egrid, self.state[spec.lidx():spec.uidx()]

    def get_solution_scale(self, nco_id, epow=0):
        """Returns the spectrum scaled back to total energy"""
        spec = self.spec_man.ncoid2sref[nco_id]
        egrid = spec.A * self.egrid
        return egrid, egrid**epow * self.state[spec.lidx():spec.uidx()] / spec.A

    def _check_id_grid(self, nco_ids, egrid):
        # Take egrid from first id ( doesn't cover the range for iron for example)
        # create a common egrid or used supplied one
        if egrid is None:
            max_mass = max([s.A for s in self.spec_man.species_refs])
            emin_log, emax_log, nbins = list(config.cosmic_ray_grid)
            emax_log = np.log10(max_mass * 10**emax_log)
            nbins *= 4
            com_egrid = EnergyGrid(emin_log, emax_log, nbins).grid
        else:
            com_egrid = egrid

        if isinstance(nco_ids, list):
            pass
        elif nco_ids == 'CR':
            nco_ids = [s for s in self.known_species if s >= 100]
        elif nco_ids == 'nu':
            nco_ids = [
                s for s in self.known_species if s in [11, 12, 13, 14, 15, 16]
            ]
        elif nco_ids == 'all':
            nco_ids = self.known_species
        elif isinstance(nco_ids, tuple):
            select, vmin, vmax = nco_ids
            nco_ids = [
                s for s in self.known_species if vmin <= select(s) <= vmax
            ]

        return nco_ids, com_egrid

    def _collect_interpolated_spectra(self, nco_ids, epow, egrid=None):
        """Collect interpolated spectra in a 2D array. Used by
        get_solution_group and get_lnA"""
        nco_ids, com_egrid = self._check_id_grid(nco_ids, egrid)
        
        # collect all the spectra in 2d array of dimension
        spectra = np.zeros((len(nco_ids), com_egrid.size))
        for idx, pid in enumerate(nco_ids):
            curr_egrid, curr_spec = self.get_solution_scale(pid, epow)
            mask = curr_spec > 0.
            if np.count_nonzero(mask) > 0:
                res = np.exp(
                    np.interp(
                        np.log(com_egrid),
                        np.log(curr_egrid[mask]),
                        np.log(curr_spec[mask]),
                        left=np.nan,
                        right=np.nan))
            else:
                res = np.zeros_like(com_egrid)
            spectra[idx] = np.nan_to_num(res)

        return nco_ids, com_egrid, spectra

    def get_solution_group(self, nco_ids, epow=3, egrid=None):
        """Return the summed spectrum (in total energy) for all elements in the range"""
        
        _, com_egrid, spectra = self._collect_interpolated_spectra(nco_ids, epow, egrid)
        spectrum = spectra.sum(axis=0)

        return com_egrid, spectrum

    def get_lnA(self, nco_ids, egrid=None):
        """Return the average ln(A) as a function of total energy for all elements in the range"""
        
        nco_ids, com_egrid, spectra = self._collect_interpolated_spectra(nco_ids, 0, egrid)

        # get the average and variance by using the spectra as weights
        lnA = np.array([np.log(self.spec_man.ncoid2sref[el].A) for el in nco_ids])
        average = (
            lnA[:, np.newaxis] * spectra).sum(axis=0) / spectra.sum(axis=0)
        variance = (lnA[:, np.newaxis]**2 *
                    spectra).sum(axis=0) / spectra.sum(axis=0) - average**2

        return com_egrid, average, variance

    def get_energy_density(self, nco_id):
        from scipy.integrate import trapz

        A = self.spec_man.ncoid2sref[nco_id].A
        return trapz(A * self.egrid * self.get_solution(nco_id), self.egrid)


class UHECRPropagationSolver(object):
    def __init__(
            self,
            initial_z,
            final_z,
            prince_run,
            enable_adiabatic_losses=True,
            enable_pairprod_losses=True,
            enable_photohad_losses=True,
            enable_injection_jacobian=True,
            enable_partial_diff_jacobian=True,
            z_offset=0.,
    ):
        self.initial_z = initial_z + z_offset
        self.final_z = final_z + z_offset
        self.z_offset = z_offset

        self.current_z_rates = None
        self.recomp_z_threshold = config.update_rates_z_threshold

        self.spec_man = prince_run.spec_man
        self.egrid = prince_run.cr_grid.grid
        self.ebins = prince_run.cr_grid.bins
        # Flags to enable/disable different loss types
        self.enable_adiabatic_losses = enable_adiabatic_losses
        self.enable_pairprod_losses = enable_pairprod_losses
        self.enable_photohad_losses = enable_photohad_losses
        # Flag for Jacobian injection:
        # if True injection in jacobion
        # if False injection only per dz-step
        self.enable_injection_jacobian = enable_injection_jacobian
        self.enable_partial_diff_jacobian = enable_partial_diff_jacobian

        self.had_int_rates = prince_run.int_rates
        self.adia_loss_rates_grid = prince_run.adia_loss_rates_grid
        self.pair_loss_rates_grid = prince_run.pair_loss_rates_grid
        self.adia_loss_rates_bins = prince_run.adia_loss_rates_bins
        self.pair_loss_rates_bins = prince_run.pair_loss_rates_bins
        self.intp = None

        self.state = np.zeros(prince_run.dim_states)
        self.result = None
        self.dim_states = prince_run.dim_states

        self.list_of_sources = []

        self.diff_operator = DifferentialOperator(prince_run.cr_grid,
                                                  prince_run.spec_man.nspec).operator
        self.semi_lag_solver = SemiLagrangianSolver(prince_run.cr_grid)

        # Configuration of BLAS backend
        self.using_cupy = False
        if config.has_cupy and config.linear_algebra_backend.lower() == 'cupy':
            self.eqn_derivative = self.eqn_deriv_cupy
            self.using_cupy = True
        elif config.has_mkl and config.linear_algebra_backend.lower() == 'mkl':
            self.eqn_derivative = self.eqn_deriv_mkl
        else:
            self.eqn_derivative = self.eqn_deriv_standard

        # Reset counters
        self.ncallsf = 0
        self.ncallsj = 0

    @property
    def known_species(self):
        return self.spec_man.known_species

    @property
    def res(self):
        if self.result is None:
            self.result = UHECRPropagationResult(self.state, self.egrid,
                                                 self.spec_man)
        return self.result

    def add_source_class(self, source_instance):
        self.list_of_sources.append(source_instance)

    def dldz(self, z):
        return -1. / ((1. + z) * H(z) * PRINCE_UNITS.cm2sec)

    def injection(self, dz, z):
        """This needs to return the injection rate
        at each redshift value z"""
        f = self.dldz(z) * dz * PRINCE_UNITS.cm2sec
        if len(self.list_of_sources) > 1:
            return f * np.sum(
                [s.injection_rate(z) for s in self.list_of_sources], axis=0)
        else:
            return f * self.list_of_sources[0].injection_rate(z)

    def _update_jacobian(self, z):
        info(5, 'Updating jacobian matrix at redshift', z)

        # enable photohadronic losses, or use a zero matrix
        if self.enable_photohad_losses:
            self.jacobian = self.had_int_rates.get_hadr_jacobian(
                z, self.dldz(z), force_update=True)
        else:
            self.jacobian = self.had_int_rates.get_hadr_jacobian(
                z, self.dldz(z), force_update=True)
            self.jacobian.data *= 0.

        self.last_hadr_jac = None

    def semi_lagrangian(self, delta_z, z, state):
        z = z - self.z_offset

        # if no continuous losses are enables, just return the state.
        if not self.enable_adiabatic_losses and not self.enable_pairprod_losses:
            return state

        conloss = np.zeros_like(self.adia_loss_rates_bins.energy_vector)
        if self.enable_adiabatic_losses:
            conloss += self.adia_loss_rates_bins.loss_vector(z)
        if self.enable_pairprod_losses:
            conloss += self.pair_loss_rates_bins.loss_vector(z)
        conloss *= self.dldz(z) * delta_z

        method = config.semi_lagr_method

        # -------------------------------------------------------------------
        # numpy interpolator
        # -------------------------------------------------------------------
        if method == 'intp_numpy':
            for spec in self.spec_man.species_refs:
                lbin, ubin = spec.lbin(), spec.ubin()
                lidx, uidx = spec.lidx(), spec.uidx()
                state[lidx:uidx] = self.semi_lag_solver.interpolate(
                    conloss[lbin:ubin], state[lidx:uidx])
        # -------------------------------------------------------------------
        # local gradient interpolation
        # -------------------------------------------------------------------
        elif method == 'gradient':
            for spec in self.spec_man.species_refs:
                lbin, ubin = spec.lbin(), spec.ubin()
                lidx, uidx = spec.lidx(), spec.uidx()
                state[lidx:uidx] = self.semi_lag_solver.interpolate_gradient(
                    conloss[lbin:ubin], state[lidx:uidx])

        # -------------------------------------------------------------------
        # linear lagrange weigts
        # -------------------------------------------------------------------
        elif method == 'linear':
            for spec in self.spec_man.species_refs:
                lbin, ubin = spec.lbin(), spec.ubin()
                lidx, uidx = spec.lidx(), spec.uidx()
                state[lidx:
                      uidx] = self.semi_lag_solver.interpolate_linear_weights(
                          conloss[lbin:ubin], state[lidx:uidx])

        # -------------------------------------------------------------------
        # quadratic lagrange weigts
        # -------------------------------------------------------------------
        elif method == 'quadratic':
            for spec in self.spec_man.species_refs:
                lbin, ubin = spec.lbin(), spec.ubin()
                lidx, uidx = spec.lidx(), spec.uidx()
                state[
                    lidx:
                    uidx] = self.semi_lag_solver.interpolate_quadratic_weights(
                        conloss[lbin:ubin], state[lidx:uidx])

        # -------------------------------------------------------------------
        # cubic lagrange weigts
        # -------------------------------------------------------------------
        elif method == 'cubic':
            for spec in self.spec_man.species_refs:
                lbin, ubin = spec.lbin(), spec.ubin()
                lidx, uidx = spec.lidx(), spec.uidx()
                state[lidx:
                      uidx] = self.semi_lag_solver.interpolate_cubic_weights(
                          conloss[lbin:ubin], state[lidx:uidx])

        # -------------------------------------------------------------------
        # 4th order lagrange weigts
        # -------------------------------------------------------------------
        elif method == '4th_order':
            for spec in self.spec_man.species_refs:
                lbin, ubin = spec.lbin(), spec.ubin()
                lidx, uidx = spec.lidx(), spec.uidx()
                state[
                    lidx:
                    uidx] = self.semi_lag_solver.interpolate_4thorder_weights(
                        conloss[lbin:ubin], state[lidx:uidx])

        # -------------------------------------------------------------------
        # 5th order lagrange weigts
        # -------------------------------------------------------------------
        elif method == '5th_order':
            for spec in self.spec_man.species_refs:
                lbin, ubin = spec.lbin(), spec.ubin()
                lidx, uidx = spec.lidx(), spec.uidx()
                state[
                    lidx:
                    uidx] = self.semi_lag_solver.interpolate_5thorder_weights(
                        conloss[lbin:ubin], state[lidx:uidx])

        # -------------------------------------------------------------------
        # finite diff euler steps
        # -------------------------------------------------------------------
        elif method == 'finite_diff':
            conloss = np.zeros_like(self.adia_loss_rates_grid.energy_vector)
            if self.enable_adiabatic_losses:
                conloss += self.adia_loss_rates_grid.loss_vector(z)
            if self.enable_pairprod_losses:
                conloss += self.pair_loss_rates_grid.loss_vector(z)

            state[:] = state + self.dldz(
                z) * delta_z * self.diff_operator.dot(conloss * state)

        # -------------------------------------------------------------------
        # if method was not in list before, raise an Expection
        # -------------------------------------------------------------------
        else:
            raise Exception(
                'Unknown semi-lagrangian method ({:})'.format(method))

        return state

    def eqn_deriv_standard(self, z, state, *args):
        z = z - self.z_offset

        self.ncallsf += 1

        # Update rates/cross sections only if solver requests to do so
        if abs(z - self.current_z_rates) > self.recomp_z_threshold:
            self._update_jacobian(z)
            self.current_z_rates = z

        r = self.jacobian.dot(state)
        
        if self.enable_injection_jacobian:
            r += self.injection(1., z)[:, np.newaxis]

        if self.enable_partial_diff_jacobian:
            conloss = np.zeros_like(self.adia_loss_rates_grid.energy_vector)
            if self.enable_adiabatic_losses:
                conloss += self.adia_loss_rates_grid.loss_vector(z)
            if self.enable_pairprod_losses:
                conloss += self.pair_loss_rates_grid.loss_vector(z)
            r += self.dldz(z) * self.diff_operator.dot(
                conloss[:, np.newaxis] * state)
        return r

    def eqn_deriv_cupy(self, z, state, *args):
        import cupy
        
        z = z - self.z_offset

        self.ncallsf += 1

        # Update rates/cross sections only if solver requests to do so
        if abs(z - self.current_z_rates) > self.recomp_z_threshold:
            self._update_jacobian(z)
            self.current_z_rates = z
        
        state = cupy.array(state)

        r = self.jacobian.dot(state)
        if self.enable_injection_jacobian:
            r += cupy.array(self.injection(1., z))[:, np.newaxis]
        if self.enable_partial_diff_jacobian:
            conloss = np.zeros_like(self.adia_loss_rates_grid.energy_vector)
            if self.enable_adiabatic_losses:
                conloss += self.adia_loss_rates_grid.loss_vector(z)
            if self.enable_pairprod_losses:
                conloss += self.pair_loss_rates_grid.loss_vector(z)
            r += self.dldz(z) * self.diff_operator.dot(
                cupy.array(conloss)[:, np.newaxis] * state)
        
        return cupy.asnumpy(r)

    def eqn_deriv_mkl(self, z, state, *args):
        from prince_cr.solvers.mkl_interface import csrmm, csrmv
        if state.shape[1] > 1:
            matrix_input = True
        else:
            matrix_input = False

        # Here starts the eqn_deriv content
        z = z - self.z_offset
        self.ncallsf += 1

        # Update rates/cross sections only if solver requests to do so
        if abs(z - self.current_z_rates) > self.recomp_z_threshold:
            self._update_jacobian(z)
            self.current_z_rates = z

        res = np.zeros(state.shape)
        
        if matrix_input:
            state = np.ascontiguousarray(state)
            res = csrmm(1., self.jacobian, state, 0., res)
        else:
            res = csrmv(1., self.jacobian, state, 0., res)
            

        if self.enable_injection_jacobian:
            res += self.injection(1., z)[:, np.newaxis]
        
        if self.enable_partial_diff_jacobian:
            conloss = np.zeros_like(self.adia_loss_rates_grid.energy_vector)
            if self.enable_adiabatic_losses:
                conloss += self.adia_loss_rates_grid.loss_vector(z)
            if self.enable_pairprod_losses:
                conloss += self.pair_loss_rates_grid.loss_vector(z)
            if matrix_input:
                res = csrmm(self.dldz(z), self.diff_operator,
                    conloss[:, np.newaxis] * state, 1., res)
            else:
                res = csrmv(self.dldz(z), self.diff_operator,
                    conloss[:, np.newaxis] * state, 1., res)

        return res

class UHECRPropagationSolverBDF(UHECRPropagationSolver):

    def __init__(self, *args, **kwargs):
        self.atol = kwargs.pop('atol', 1e40)
        self.rtol = kwargs.pop('rtol', 1e-10)
        super(UHECRPropagationSolverBDF, self).__init__(*args, **kwargs)

    def _init_solver(self, dz):
        initial_state = np.zeros(self.dim_states)

        self._update_jacobian(self.initial_z)
        self.current_z_rates = self.initial_z

        # find the maximum injection and reduce the system by this
        self.red_idx = np.nonzero(self.injection(1., 0.))[0].max()
        
        # Convert csr_matrix from GPU to scipy
        try:
            sparsity = self.had_int_rates.get_hadr_jacobian(
                self.initial_z, 1.).get()
        except AttributeError:
            sparsity = self.had_int_rates.get_hadr_jacobian(
                self.initial_z, 1.)

        from prince_cr.util import PrinceBDF
        self.r = PrinceBDF(
            self.eqn_derivative,
            self.initial_z,
            initial_state,
            self.final_z,
            max_step=np.abs(dz),
            atol=self.atol,
            rtol=self.rtol,
            #  jac = self.eqn_jac,
            jac_sparsity=sparsity,
            vectorized=True)

    def solve(self,
              dz=1e-3,
              verbose=True,
              summary=False,
              extended_output=False,
              full_reset=False,
              progressbar=False):
        from time import time
        from prince_cr.util import PrinceProgressBar

        reset_counter = 0
        stepcount = 0
        dz = -1 * dz
        start_time = time()

        info(2, 'Setting up Solver')
        self._init_solver(dz)
        info(2, 'Solver initialized in {0} s'.format(time() - start_time))

        info(2, 'Starting integration.')
        with PrinceProgressBar(
            bar_type=progressbar,
            nsteps=-(self.initial_z - self.final_z) / dz) as pbar:
            while self.r.status == 'running':
                # print '------ at', self.r.t, '------'
                if verbose:
                    info(3, "Integrating at z = {0}".format(self.r.t))
                # --------------------------------------------
                # Evaluate a step
                # --------------------------------------------
                self.r.step()
                # --------------------------------------------
                # Some last checks and resets
                # --------------------------------------------
                if verbose:
                    print('last step:', self.r.step_size)
                    print('h_abs:', self.r.h_abs)
                    print('LU decomp:', self.r.nlu)
                    print('current order:', self.r.dense_output().order)
                    print('---' * 20)

                stepcount += 1
                reset_counter += 1
                pbar.update()

        if self.r.status == 'failed':
            raise Exception(
                'Integrator failed at t = {:}, try adjusting the tolerances'.
                format(self.r.t))
        if verbose:
            print('Integrator finished with t = {:}, last step was dt = {:}'.format(
                self.r.t, self.r.step_size))

        # after each run we delete the solver to save memory
        self.state = self.r.y.copy()

        if summary or verbose:
            print('Summary:')
            print('---------------------')
            print('status:   ', self.r.status)
            print('time:     ', self.r.t)
            print('last step:', self.r.step_size)
            print('RHS eval: ', self.r.nfev)
            print('Jac eval: ', self.r.njev)
            print('LU decomp:', self.r.nlu)
            print('---------------------')

        del self.r

        end_time = time()
        info(2, 'Integration completed in {0} s'.format(end_time - start_time))

class UHECRPropagationSolverEULER(UHECRPropagationSolver):
    def __init__(self, *args, **kwargs):
        super(UHECRPropagationSolverEULER, self).__init__(*args, **kwargs)

    def solve(self, dz=1e-3, verbose=True, extended_output=False,
              initial_inj=False, disable_inj=False, progressbar=False):
        from prince_cr.util import PrinceProgressBar
        from time import time
        self._update_jacobian(self.initial_z)
        self.current_z_rates = self.initial_z

        dz = -1 * dz
        start_time = time()
        curr_z = self.initial_z
        if initial_inj:
            initial_state = np.atleast_2d(self.injection(dz, self.initial_z))
            print(initial_state)
        else:
            initial_state = np.zeros((self.dim_states,1))
        state = initial_state
        info(2, 'Starting integration.')
        with PrinceProgressBar(
            bar_type=progressbar,
            nsteps=-(self.initial_z - self.final_z) / dz) as pbar:
            while curr_z + dz >= self.final_z:
                if verbose:
                    info(3, "Integrating at z={0}".format(curr_z))

                # --------------------------------------------
                # Solve for hadronic interactions
                # --------------------------------------------
                if verbose:
                    print('Solving hadr losses at t =', curr_z)
                self._update_jacobian(curr_z)

                state += self.eqn_derivative(curr_z, state) * dz

                # --------------------------------------------
                # Apply the injection
                # --------------------------------------------
                if not disable_inj:
                    if not self.enable_injection_jacobian and not self.enable_partial_diff_jacobian:
                        if verbose:
                            print('applying injection at t =', curr_z)
                        state += self.injection(dz, curr_z)

                # --------------------------------------------
                # Apply the semi lagrangian
                # --------------------------------------------
                if not self.enable_partial_diff_jacobian:
                    if self.enable_adiabatic_losses or self.enable_pairprod_losses:
                        if verbose:
                            print('applying semi lagrangian at t =', curr_z)
                        state = self.semi_lagrangian(dz, curr_z, state)

                # --------------------------------------------
                # Some last checks and resets
                # --------------------------------------------
                if curr_z < -1 * dz:
                    print('break at z =', curr_z)
                    break
                curr_z += dz
                pbar.update()

        end_time = time()
        info(2, 'Integration completed in {0} s'.format(end_time - start_time))
        self.state = state


# -*- coding: utf-8 -*-
import sys
from distutils.core import setup
from os.path import abspath, dirname, join

import setuptools

def get_version():
    version = {}
    with open("prince_cr/version.py") as fp:
        exec (fp.read(), version)
    return version['__version__']

__version__ = get_version()

this_directory = abspath(dirname(__file__))
if sys.version_info.major == 3:
    with open(join(this_directory, 'README.md'), encoding='utf-8') as f:
        long_description = f.read()
else:
    with open(join(this_directory, 'README.md')) as f:
        long_description = f.read()

skip_marker = "# PriNCe"
long_description = long_description[long_description.index(skip_marker) :].lstrip()

setup(
    name='prince_cr',
    version=__version__,
    author='Jonas Heinze and Anatoli Fedynitch',
    author_email='afedynitch@gmail.com',
    description='Cosmic ray PRopagation Including Nuclear Cascade equations',
    long_description=long_description,
    long_description_content_type='text/markdown',
    license='BSD 3-Clause License',
    url='https://github.com/joheinze/PriNCe',
    packages=[
        'prince_cr', 'prince_cr.cross_sections',
        'prince_cr.solvers', 'prince_cr.tests'
    ],
    install_requires=[
        # 'six',
        'scipy',
        'numpy',
        'tqdm',
        'h5py',
        'requests'
    ],
    tests_require=['pytest','matplotlib'],
    py_modules=["six"],
    package_data={'prince_cr': ['data/particle_data.ppo']},
    classifiers=[
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Scientific/Engineering :: Physics',
        'Intended Audience :: Science/Research',
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: BSD License'
    ]
)
